from typing import Optional
import logging
import os
import cProfile, pstats

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.profilers.operation_profiler import OperationProfiler

logger = logging.getLogger('graphsignal')

class PythonProfiler(OperationProfiler):
    def __init__(self):
        self._profiler = None
        self._exclude_path = os.path.dirname(os.path.realpath(graphsignal.__file__))

    def read_info(self, signal):
        pass

    def start(self, signal):
        logger.debug('Activating Python profiler')

        self._profiler = cProfile.Profile()
        self._profiler.enable()

    def stop(self, signal):
        logger.debug('Deactivating Python profiler')

        self._profiler.disable()
        self._convert_to_operations(signal)
        self._profiler = None

    def _convert_to_operations(self, signal):
        stats = pstats.Stats(self._profiler)
        func_list = stats.fcn_list[:] if stats.fcn_list else list(stats.stats.keys())
        if not func_list:
            return

        for func in func_list:
            file_name, line_number, func_name = func
            if file_name.startswith(self._exclude_path):
                continue
            cc, nc, tt, ct, callers = stats.stats[func]

            op_stats = signal.op_stats.add()
            op_stats.device_type = signals_pb2.DeviceType.CPU
            op_stats.op_name = '{0} ({1}:{2})'.format(func_name, file_name, line_number)
            op_stats.count = int(nc)
            op_stats.total_host_time_us = _to_us(ct)
            op_stats.self_host_time_us = _to_us(tt)


def _to_us(sec):
    return int(sec * 1e6)



import unittest
import logging
import sys
import os
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')


class PyTorchLightningTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    @patch.object(Uploader, 'upload_signal')
    def test_callback(self, mocked_upload_signal):
        import torch
        from pytorch_lightning import LightningModule, Trainer
        from torch import nn
        from torch.nn import functional as F
        from torch.utils.data import DataLoader, random_split
        from torchmetrics import Accuracy
        from torchvision import transforms
        from torchvision.datasets import MNIST
        from graphsignal.callbacks.pytorch_lightning import GraphsignalCallback

        PATH_DATASETS = os.environ.get("PATH_DATASETS", ".")
        AVAIL_GPUS = min(1, torch.cuda.device_count())
        BATCH_SIZE = 256 if AVAIL_GPUS else 64

        class MNISTModel(LightningModule):
            def __init__(self):
                super().__init__()
                self.batch_size = BATCH_SIZE
                self.l1 = torch.nn.Linear(28 * 28, 10)
                self.train_acc = Accuracy()
                self.test_acc = Accuracy()

            def forward(self, x):
                return torch.relu(self.l1(x.view(x.size(0), -1)))

            def training_step(self, batch, batch_nb):
                x, y = batch
                preds = self(x)
                loss = F.cross_entropy(preds, y)
                self.train_acc(preds, y)
                self.log('train_acc', self.train_acc, on_step=True, on_epoch=False)
                return loss

            def test_step(self, batch, batch_nb):
                x, y = batch
                preds = self(x)
                loss = F.cross_entropy(preds, y)
                self.test_acc(preds, y)
                self.log('test_acc', self.test_acc, on_step=True, on_epoch=False)
                return loss

            def train_dataloader(self):
                train_ds = MNIST(PATH_DATASETS, train=True, download=True, transform=transforms.ToTensor())
                train_ds = torch.utils.data.Subset(train_ds, torch.arange(1000))
                train_loader = DataLoader(train_ds, batch_size=self.batch_size)
                return train_loader

            def test_dataloader(self):
                test_ds = MNIST(PATH_DATASETS, train=False, download=True, transform=transforms.ToTensor())
                test_ds = torch.utils.data.Subset(test_ds, torch.arange(1000))
                test_loader = DataLoader(test_ds, batch_size=self.batch_size)
                return test_loader

            def configure_optimizers(self):
                return torch.optim.Adam(self.parameters(), lr=0.02)

        mnist_model = MNISTModel()

        trainer = Trainer(
            accelerator='gpu' if torch.cuda.is_available() else 'cpu',
            devices=AVAIL_GPUS,
            max_epochs=1,
            callbacks=[GraphsignalCallback(endpoint='ep1')]
        )

        trainer.tune(mnist_model)

        trainer.fit(mnist_model)

        trainer.test(mnist_model)

        signal = mocked_upload_signal.call_args[0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(signal))

        self.assertEqual(signal.endpoint, 'ep1')

        self.assertTrue(
            signal.agent_info.framework_profiler_type, 
            signals_pb2.AgentInfo.ProfilerType.PYTORCH_LIGHTNING_PROFILER)

        self.assertEqual(
            signal.frameworks[-1].type,
            signals_pb2.FrameworkInfo.FrameworkType.PYTORCH_LIGHTNING_FRAMEWORK)
        self.assertTrue(signal.frameworks[-1].version.major > 0)

        self.assertEqual(
            signal.model_info.model_format,
            signals_pb2.ModelInfo.ModelFormat.PYTORCH_FORMAT)
        self.assertTrue(signal.model_info.model_size_bytes > 0)

        test_op_stats = None
        for op_stats in signal.op_stats:
            if op_stats.op_name == 'aten::addmm':
                test_op_stats = op_stats
                break
        self.assertIsNotNone(test_op_stats)
        self.assertTrue(test_op_stats.count >= 1)
        import torch
        if torch.cuda.is_available():
            self.assertTrue(test_op_stats.total_device_time_us >= 1)
            self.assertTrue(test_op_stats.self_device_time_us >= 1)
        else:
            self.assertTrue(test_op_stats.total_host_time_us >= 1)
            self.assertTrue(test_op_stats.self_host_time_us >= 1)



import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.profilers.python import PythonProfiler
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')


class PythonProfilerTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_stat_stop(self):
        def slow_method():
            time.sleep(0.1)

        profiler = PythonProfiler()
        profiler._exclude_path = 'donotmatchpath'
        signal = signals_pb2.WorkerSignal()
        profiler.start(signal)
        slow_method()
        slow_method()
        profiler.stop(signal)

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(signal))

        foundOp = False
        for op_stats in signal.op_stats:
            if op_stats.op_name.startswith('slow_method') and op_stats.count == 2 and op_stats.total_host_time_us > 200000:
                foundOp = True
                break
        self.assertTrue(foundOp)



import logging
import os
import tempfile
import gzip
import shutil
import glob

from graphsignal.profilers.tensorflow_proto import overview_page_pb2
from graphsignal.profilers.tensorflow_proto import input_pipeline_pb2
from graphsignal.profilers.tensorflow_proto import tf_stats_pb2
from graphsignal.profilers.tensorflow_proto import kernel_stats_pb2
from graphsignal.profilers.tensorflow_proto import memory_profile_pb2

from graphsignal.proto import signals_pb2

logger = logging.getLogger('graphsignal')

def create_log_dir():
    log_dir = tempfile.mkdtemp(prefix='graphsignal-')
    logger.debug('Created temporary log directory %s', log_dir)
    return log_dir

def remove_log_dir(log_dir):
    shutil.rmtree(log_dir)
    logger.debug('Removed temporary log directory %s', log_dir)

def find_and_read(log_dir, file_pattern, decompress=True, max_size=None):
    file_paths = glob.glob(os.path.join(log_dir, file_pattern))
    if len(file_paths) == 0:
        logger.debug('Files are not found at %s', os.path.join(log_dir, file_pattern))
        return None

    found_path = file_paths[-1]

    if max_size:
        file_size = os.path.getsize(found_path)
        if file_size > max_size:
            raise Exception('File is too big: {0}'.format(file_size))

    if decompress and found_path.endswith('.gz'):
        last_file = gzip.open(found_path, "rb")
    else:
        last_file = open(found_path, "rb")
    data = last_file.read()
    last_file.close()

    return data

def convert_tensorflow_profile(log_dir, signal):
    # Read trace
    trace_json_gz = find_and_read(
        log_dir,
        'plugins/profile/*/*trace.json.gz',
        decompress=False,
        max_size=5 * 1e6)
    if trace_json_gz:
        signal.trace_data = trace_json_gz

    # Operation stats
    tf_stats_data = find_and_read(
        log_dir,
        'plugins/profile/*/*tensorflow_stats.pb')
    if tf_stats_data:
        tf_stats_db = tf_stats_pb2.TfStatsDatabase()
        tf_stats_db.ParseFromString(tf_stats_data)
        for tf_stats_record in tf_stats_db.without_idle.tf_stats_record:
            op_stats = signal.op_stats.add()
            if tf_stats_record.host_or_device == 'Host':
                op_stats.device_type = signals_pb2.DeviceType.CPU
                op_stats.total_host_time_us = _uint(tf_stats_record.total_time_in_us)
                op_stats.self_host_time_us = _uint(tf_stats_record.total_self_time_in_us)
                op_stats.self_host_memory_rate = _uint(tf_stats_record.measured_memory_bw)
            else:
                op_stats.device_type = signals_pb2.DeviceType.GPU
                op_stats.total_device_time_us = _uint(tf_stats_record.total_time_in_us)
                op_stats.self_device_time_us = _uint(tf_stats_record.total_self_time_in_us)
                op_stats.self_device_memory_rate = _uint(tf_stats_record.measured_memory_bw)
                op_stats.tensorcore_utilization = tf_stats_record.gpu_tensorcore_utilization
            op_stats.op_type = tf_stats_record.op_type
            op_stats.op_name = tf_stats_record.op_name
            op_stats.count = _uint(tf_stats_record.occurrences)
            op_stats.flops_per_sec = _uint(tf_stats_record.measured_flop_rate)
    else:
        logger.debug('No operation data found in log directory')

    # Kernel stats
    kernel_stats_data = find_and_read(
        log_dir,
        'plugins/profile/*/*kernel_stats.pb')
    if kernel_stats_data:
        kernel_stats_db = kernel_stats_pb2.KernelStatsDb()
        kernel_stats_db.ParseFromString(kernel_stats_data)
        for kernel_report in kernel_stats_db.reports:
            kernel_stats = signal.kernel_stats.add()
            kernel_stats.device_type = signals_pb2.DeviceType.GPU
            kernel_stats.op_name = kernel_report.op_name
            kernel_stats.kernel_name = kernel_report.name
            kernel_stats.count = _uint(kernel_report.occurrences)
            kernel_stats.duration_ns = _uint(kernel_report.total_duration_ns)
            kernel_stats.is_using_tensorcore = kernel_report.is_kernel_using_tensor_core
    else:
        logger.debug('No kernel data found in log directory')

def _uint(val):
    return max(int(val), 0)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: graphsignal/profilers/tensorflow_proto/tf_stats.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n5graphsignal/profilers/tensorflow_proto/tf_stats.proto\x12\x13tensorflow.profiler\"\xa7\x01\n\x0fTfStatsDatabase\x12\x34\n\twith_idle\x18\x04 \x01(\x0b\x32!.tensorflow.profiler.TfStatsTable\x12\x37\n\x0cwithout_idle\x18\x05 \x01(\x0b\x32!.tensorflow.profiler.TfStatsTable\x12\x13\n\x0b\x64\x65vice_type\x18\x06 \x01(\tJ\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03J\x04\x08\x03\x10\x04\"\x83\x01\n\x0cTfStatsTable\x12;\n\x0ftf_stats_record\x18\x01 \x03(\x0b\x32\".tensorflow.profiler.TfStatsRecord\x12\x19\n\x11host_tf_pprof_key\x18\x02 \x01(\t\x12\x1b\n\x13\x64\x65vice_tf_pprof_key\x18\x03 \x01(\t\"\xbb\x04\n\rTfStatsRecord\x12\x0c\n\x04rank\x18\x01 \x01(\x04\x12\x16\n\x0ehost_or_device\x18\x02 \x01(\t\x12\x0f\n\x07op_type\x18\x03 \x01(\t\x12\x0f\n\x07op_name\x18\x04 \x01(\t\x12\x13\n\x0boccurrences\x18\x05 \x01(\x03\x12\x18\n\x10total_time_in_us\x18\x06 \x01(\x01\x12\x16\n\x0e\x61vg_time_in_us\x18\x07 \x01(\x01\x12\x1d\n\x15total_self_time_in_us\x18\x08 \x01(\x01\x12\x1b\n\x13\x61vg_self_time_in_us\x18\t \x01(\x01\x12*\n\"device_total_self_time_as_fraction\x18\n \x01(\x01\x12\x35\n-device_cumulative_total_self_time_as_fraction\x18\x0b \x01(\x01\x12(\n host_total_self_time_as_fraction\x18\x0c \x01(\x01\x12\x33\n+host_cumulative_total_self_time_as_fraction\x18\r \x01(\x01\x12\x1a\n\x12measured_flop_rate\x18\x0e \x01(\x01\x12\x1a\n\x12measured_memory_bw\x18\x0f \x01(\x01\x12\x1d\n\x15operational_intensity\x18\x10 \x01(\x01\x12\x10\n\x08\x62ound_by\x18\x11 \x01(\t\x12\x10\n\x08is_eager\x18\x12 \x01(\x08\x12\"\n\x1agpu_tensorcore_utilization\x18\x13 \x01(\x01\x62\x06proto3')



_TFSTATSDATABASE = DESCRIPTOR.message_types_by_name['TfStatsDatabase']
_TFSTATSTABLE = DESCRIPTOR.message_types_by_name['TfStatsTable']
_TFSTATSRECORD = DESCRIPTOR.message_types_by_name['TfStatsRecord']
TfStatsDatabase = _reflection.GeneratedProtocolMessageType('TfStatsDatabase', (_message.Message,), {
  'DESCRIPTOR' : _TFSTATSDATABASE,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.tf_stats_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.TfStatsDatabase)
  })
_sym_db.RegisterMessage(TfStatsDatabase)

TfStatsTable = _reflection.GeneratedProtocolMessageType('TfStatsTable', (_message.Message,), {
  'DESCRIPTOR' : _TFSTATSTABLE,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.tf_stats_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.TfStatsTable)
  })
_sym_db.RegisterMessage(TfStatsTable)

TfStatsRecord = _reflection.GeneratedProtocolMessageType('TfStatsRecord', (_message.Message,), {
  'DESCRIPTOR' : _TFSTATSRECORD,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.tf_stats_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.TfStatsRecord)
  })
_sym_db.RegisterMessage(TfStatsRecord)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TFSTATSDATABASE._serialized_start=79
  _TFSTATSDATABASE._serialized_end=246
  _TFSTATSTABLE._serialized_start=249
  _TFSTATSTABLE._serialized_end=380
  _TFSTATSRECORD._serialized_start=383
  _TFSTATSRECORD._serialized_end=954
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: graphsignal/profilers/tensorflow_proto/memory_profile.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n;graphsignal/profilers/tensorflow_proto/memory_profile.proto\x12\x13tensorflow.profiler\"\xa1\x01\n\x16MemoryAggregationStats\x12\x1c\n\x14stack_reserved_bytes\x18\x01 \x01(\x03\x12\x1c\n\x14heap_allocated_bytes\x18\x02 \x01(\x03\x12\x19\n\x11\x66ree_memory_bytes\x18\x03 \x01(\x03\x12\x15\n\rfragmentation\x18\x04 \x01(\x01\x12\x19\n\x11peak_bytes_in_use\x18\x05 \x01(\x03\"\xfd\x01\n\x16MemoryActivityMetadata\x12<\n\x0fmemory_activity\x18\x01 \x01(\x0e\x32#.tensorflow.profiler.MemoryActivity\x12\x17\n\x0frequested_bytes\x18\x02 \x01(\x03\x12\x18\n\x10\x61llocation_bytes\x18\x03 \x01(\x03\x12\x0f\n\x07\x61\x64\x64ress\x18\x04 \x01(\x04\x12\x12\n\ntf_op_name\x18\x05 \x01(\t\x12\x0f\n\x07step_id\x18\x06 \x01(\x03\x12\x13\n\x0bregion_type\x18\x07 \x01(\t\x12\x11\n\tdata_type\x18\x08 \x01(\t\x12\x14\n\x0ctensor_shape\x18\t \x01(\t\"\xbf\x01\n\x15MemoryProfileSnapshot\x12\x16\n\x0etime_offset_ps\x18\x01 \x01(\x03\x12\x46\n\x11\x61ggregation_stats\x18\x02 \x01(\x0b\x32+.tensorflow.profiler.MemoryAggregationStats\x12\x46\n\x11\x61\x63tivity_metadata\x18\x03 \x01(\x0b\x32+.tensorflow.profiler.MemoryActivityMetadata\"\xaf\x01\n\x14MemoryProfileSummary\x12!\n\x19peak_bytes_usage_lifetime\x18\x01 \x01(\x03\x12?\n\npeak_stats\x18\x02 \x01(\x0b\x32+.tensorflow.profiler.MemoryAggregationStats\x12\x1a\n\x12peak_stats_time_ps\x18\x03 \x01(\x03\x12\x17\n\x0fmemory_capacity\x18\x04 \x01(\x03\"Z\n\x10\x41\x63tiveAllocation\x12\x16\n\x0esnapshot_index\x18\x01 \x01(\x03\x12\x15\n\rspecial_index\x18\x02 \x01(\x03\x12\x17\n\x0fnum_occurrences\x18\x03 \x01(\x03\"\xba\x02\n\x19PerAllocatorMemoryProfile\x12L\n\x18memory_profile_snapshots\x18\x01 \x03(\x0b\x32*.tensorflow.profiler.MemoryProfileSnapshot\x12\x42\n\x0fprofile_summary\x18\x02 \x01(\x0b\x32).tensorflow.profiler.MemoryProfileSummary\x12\x41\n\x12\x61\x63tive_allocations\x18\x03 \x03(\x0b\x32%.tensorflow.profiler.ActiveAllocation\x12H\n\x13special_allocations\x18\x04 \x03(\x0b\x32+.tensorflow.profiler.MemoryActivityMetadata\"\x97\x02\n\rMemoryProfile\x12g\n\x1cmemory_profile_per_allocator\x18\x01 \x03(\x0b\x32\x41.tensorflow.profiler.MemoryProfile.MemoryProfilePerAllocatorEntry\x12\x11\n\tnum_hosts\x18\x02 \x01(\x05\x12\x12\n\nmemory_ids\x18\x03 \x03(\t\x1ap\n\x1eMemoryProfilePerAllocatorEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12=\n\x05value\x18\x02 \x01(\x0b\x32..tensorflow.profiler.PerAllocatorMemoryProfile:\x02\x38\x01J\x04\x08\x04\x10\x05*h\n\x0eMemoryActivity\x12\x14\n\x10UNKNOWN_ACTIVITY\x10\x00\x12\x0e\n\nALLOCATION\x10\x01\x12\x10\n\x0c\x44\x45\x41LLOCATION\x10\x02\x12\x0f\n\x0bRESERVATION\x10\x03\x12\r\n\tEXPANSION\x10\x04\x62\x06proto3')

_MEMORYACTIVITY = DESCRIPTOR.enum_types_by_name['MemoryActivity']
MemoryActivity = enum_type_wrapper.EnumTypeWrapper(_MEMORYACTIVITY)
UNKNOWN_ACTIVITY = 0
ALLOCATION = 1
DEALLOCATION = 2
RESERVATION = 3
EXPANSION = 4


_MEMORYAGGREGATIONSTATS = DESCRIPTOR.message_types_by_name['MemoryAggregationStats']
_MEMORYACTIVITYMETADATA = DESCRIPTOR.message_types_by_name['MemoryActivityMetadata']
_MEMORYPROFILESNAPSHOT = DESCRIPTOR.message_types_by_name['MemoryProfileSnapshot']
_MEMORYPROFILESUMMARY = DESCRIPTOR.message_types_by_name['MemoryProfileSummary']
_ACTIVEALLOCATION = DESCRIPTOR.message_types_by_name['ActiveAllocation']
_PERALLOCATORMEMORYPROFILE = DESCRIPTOR.message_types_by_name['PerAllocatorMemoryProfile']
_MEMORYPROFILE = DESCRIPTOR.message_types_by_name['MemoryProfile']
_MEMORYPROFILE_MEMORYPROFILEPERALLOCATORENTRY = _MEMORYPROFILE.nested_types_by_name['MemoryProfilePerAllocatorEntry']
MemoryAggregationStats = _reflection.GeneratedProtocolMessageType('MemoryAggregationStats', (_message.Message,), {
  'DESCRIPTOR' : _MEMORYAGGREGATIONSTATS,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.memory_profile_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.MemoryAggregationStats)
  })
_sym_db.RegisterMessage(MemoryAggregationStats)

MemoryActivityMetadata = _reflection.GeneratedProtocolMessageType('MemoryActivityMetadata', (_message.Message,), {
  'DESCRIPTOR' : _MEMORYACTIVITYMETADATA,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.memory_profile_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.MemoryActivityMetadata)
  })
_sym_db.RegisterMessage(MemoryActivityMetadata)

MemoryProfileSnapshot = _reflection.GeneratedProtocolMessageType('MemoryProfileSnapshot', (_message.Message,), {
  'DESCRIPTOR' : _MEMORYPROFILESNAPSHOT,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.memory_profile_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.MemoryProfileSnapshot)
  })
_sym_db.RegisterMessage(MemoryProfileSnapshot)

MemoryProfileSummary = _reflection.GeneratedProtocolMessageType('MemoryProfileSummary', (_message.Message,), {
  'DESCRIPTOR' : _MEMORYPROFILESUMMARY,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.memory_profile_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.MemoryProfileSummary)
  })
_sym_db.RegisterMessage(MemoryProfileSummary)

ActiveAllocation = _reflection.GeneratedProtocolMessageType('ActiveAllocation', (_message.Message,), {
  'DESCRIPTOR' : _ACTIVEALLOCATION,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.memory_profile_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.ActiveAllocation)
  })
_sym_db.RegisterMessage(ActiveAllocation)

PerAllocatorMemoryProfile = _reflection.GeneratedProtocolMessageType('PerAllocatorMemoryProfile', (_message.Message,), {
  'DESCRIPTOR' : _PERALLOCATORMEMORYPROFILE,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.memory_profile_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.PerAllocatorMemoryProfile)
  })
_sym_db.RegisterMessage(PerAllocatorMemoryProfile)

MemoryProfile = _reflection.GeneratedProtocolMessageType('MemoryProfile', (_message.Message,), {

  'MemoryProfilePerAllocatorEntry' : _reflection.GeneratedProtocolMessageType('MemoryProfilePerAllocatorEntry', (_message.Message,), {
    'DESCRIPTOR' : _MEMORYPROFILE_MEMORYPROFILEPERALLOCATORENTRY,
    '__module__' : 'graphsignal.profilers.tensorflow_proto.memory_profile_pb2'
    # @@protoc_insertion_point(class_scope:tensorflow.profiler.MemoryProfile.MemoryProfilePerAllocatorEntry)
    })
  ,
  'DESCRIPTOR' : _MEMORYPROFILE,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.memory_profile_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.MemoryProfile)
  })
_sym_db.RegisterMessage(MemoryProfile)
_sym_db.RegisterMessage(MemoryProfile.MemoryProfilePerAllocatorEntry)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MEMORYPROFILE_MEMORYPROFILEPERALLOCATORENTRY._options = None
  _MEMORYPROFILE_MEMORYPROFILEPERALLOCATORENTRY._serialized_options = b'8\001'
  _MEMORYACTIVITY._serialized_start=1567
  _MEMORYACTIVITY._serialized_end=1671
  _MEMORYAGGREGATIONSTATS._serialized_start=85
  _MEMORYAGGREGATIONSTATS._serialized_end=246
  _MEMORYACTIVITYMETADATA._serialized_start=249
  _MEMORYACTIVITYMETADATA._serialized_end=502
  _MEMORYPROFILESNAPSHOT._serialized_start=505
  _MEMORYPROFILESNAPSHOT._serialized_end=696
  _MEMORYPROFILESUMMARY._serialized_start=699
  _MEMORYPROFILESUMMARY._serialized_end=874
  _ACTIVEALLOCATION._serialized_start=876
  _ACTIVEALLOCATION._serialized_end=966
  _PERALLOCATORMEMORYPROFILE._serialized_start=969
  _PERALLOCATORMEMORYPROFILE._serialized_end=1283
  _MEMORYPROFILE._serialized_start=1286
  _MEMORYPROFILE._serialized_end=1565
  _MEMORYPROFILE_MEMORYPROFILEPERALLOCATORENTRY._serialized_start=1447
  _MEMORYPROFILE_MEMORYPROFILEPERALLOCATORENTRY._serialized_end=1559
# @@protoc_insertion_point(module_scope)



import unittest
import logging
import sys
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')


class KerasCallbackTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    @patch.object(Uploader, 'upload_signal')
    def test_callback(self, mocked_upload_signal):
        import tensorflow as tf
        import tensorflow_datasets as tfds
        tfds.disable_progress_bar()

        (ds_train, ds_test), ds_info = tfds.load(
            'mnist',
            split=['train', 'test'],
            shuffle_files=True,
            as_supervised=True,
            with_info=True,
        )

        def normalize_img(image, label):
            """Normalizes images: `uint8` -> `float32`."""
            return tf.cast(image, tf.float32) / 255., label

        ds_train = ds_train.map(normalize_img)
        ds_train = ds_train.batch(128)
        ds_train = ds_train.take(100)
        ds_train = ds_train.cache()
        ds_train = ds_train.prefetch(tf.data.experimental.AUTOTUNE)

        ds_test = ds_test.map(normalize_img)
        ds_test = ds_test.batch(128)
        ds_test = ds_test.take(100)
        ds_test = ds_test.cache()
        ds_test = ds_test.prefetch(tf.data.experimental.AUTOTUNE)

        model = tf.keras.models.Sequential([
            tf.keras.layers.Flatten(input_shape=(28, 28, 1)),
            tf.keras.layers.Dense(128, activation='relu'),
            tf.keras.layers.Dense(10, activation='softmax')
        ])
        model.compile(
            loss='sparse_categorical_crossentropy',
            optimizer=tf.keras.optimizers.Adam(0.001),
            metrics=['accuracy']
        )

        from graphsignal.callbacks.keras import GraphsignalCallback

        model.fit(ds_train,
            batch_size=128,
            epochs=1,
            validation_data=ds_test)

        model.evaluate(ds_test,
            batch_size=128,
            callbacks=[GraphsignalCallback(endpoint='ep1')])

        signal = mocked_upload_signal.call_args[0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(signal))

        self.assertEqual(signal.endpoint, 'ep1')

        test_op_stats = None
        for op_stats in signal.op_stats:
            if 'MatMul' in op_stats.op_type:
                test_op_stats = op_stats
                break
        self.assertIsNotNone(test_op_stats)
        self.assertTrue(test_op_stats.count >= 1)
        if len(tf.config.list_physical_devices('GPU')) > 0:
            self.assertTrue(test_op_stats.total_device_time_us >= 1)
            self.assertTrue(test_op_stats.self_device_time_us >= 1)
        else:
            self.assertTrue(test_op_stats.total_host_time_us >= 1)
            self.assertTrue(test_op_stats.self_host_time_us >= 1)



import logging
import time
import random
from threading import Lock

logger = logging.getLogger('graphsignal')

# global sampling lock
_sampling_lock = Lock()

class TraceSampler:
    MAX_ENSURED_SPANS = 2
    PREDEFINED_SPANS = [1, 10, 100, 1000]
    MIN_INTERVAL_SEC = 60

    def __init__(self):
        self._current_span = 0
        self._ensured_span_count = 0
        self._start_ts = time.time()
        self._last_sample_ts = None
        self._span_filter = {span for span in TraceSampler.PREDEFINED_SPANS}
        self._interval_mode = False
        self._profiled_count = 0

    def lock(self, ensure=False):
        self._current_span += 1

        if _sampling_lock.locked():
            return False

        now = time.time()

        if not self._interval_mode and now - self._start_ts > self.MIN_INTERVAL_SEC:
            # switch to interval-based sampling and sample first interval
            self._interval_mode = True
            _sampling_lock.acquire(blocking=False)
            return True

        if ensure:
            self._ensured_span_count += 1
            if self._ensured_span_count <= TraceSampler.MAX_ENSURED_SPANS:
                _sampling_lock.acquire(blocking=False)
                return True

        if not self._interval_mode:
            if self._current_span in self._span_filter:
                _sampling_lock.acquire(blocking=False)
                return True
        else:
            if not self._last_sample_ts:
                last_interval_sec = now - self._start_ts
            else:
                last_interval_sec = now - self._last_sample_ts
            if last_interval_sec > self.MIN_INTERVAL_SEC:
                self._ensured_span_count = 0
                _sampling_lock.acquire(blocking=False)
                return True

        return False

    def unlock(self):
        if not _sampling_lock.locked():
            return
        self._last_sample_ts = time.time()
        _sampling_lock.release()

    def should_profile(self):
        # only profile few initial spans
        if not self._interval_mode and self._current_span in self._span_filter:
            return True
        return False


import unittest
import logging
import sys
import os
import json
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')


class JaxProfilerTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_read_info(self):
        try:
            import jax
            import jax.numpy as jnp
            from jax import grad, jit, vmap
            from jax import random
        except ImportError:
            logger.info('Not testing JAX profiler, package not found.')
            return

        from graphsignal.profilers.jax import JaxProfiler

        profiler = JaxProfiler()
        signal = signals_pb2.WorkerSignal()
        profiler.read_info(signal)

        self.assertEqual(
            signal.frameworks[0].type,
            signals_pb2.FrameworkInfo.FrameworkType.JAX_FRAMEWORK)

    def test_start_stop(self):
        try:
            import jax
            import jax.numpy as jnp
            from jax import grad, jit, vmap
            from jax import random
        except ImportError:
            logger.info('Not testing JAX profiler, package not found.')
            return

        from graphsignal.profilers.jax import JaxProfiler

        profiler = JaxProfiler()
        signal = signals_pb2.WorkerSignal()
        profiler.start(signal)

        key = random.PRNGKey(0)
        x = random.normal(key, (10,))
        size = 100
        x = random.normal(key, (size, size), dtype=jnp.float32)
        jnp.dot(x, x.T).block_until_ready()

        profiler.stop(signal)

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(signal))

        test_op_stats = None
        for op_stats in signal.op_stats:
            if op_stats.op_name == 'Thunk':
                test_op_stats = op_stats
                break
        self.assertIsNotNone(test_op_stats)
        self.assertTrue(test_op_stats.count >= 1)
        if jax.device_count() > 0:
            self.assertTrue(test_op_stats.total_device_time_us >= 1)
            self.assertTrue(test_op_stats.self_device_time_us >= 1)
        else:
            self.assertTrue(test_op_stats.total_host_time_us >= 1)
            self.assertTrue(test_op_stats.self_host_time_us >= 1)

        self.assertNotEqual(signal.trace_data, b'')



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: signals.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rsignals.proto\x12\x13graphsignal.signals\"]\n\rUploadRequest\x12\x39\n\x0eworker_signals\x18\x01 \x03(\x0b\x32!.graphsignal.signals.WorkerSignal\x12\x11\n\tupload_ms\x18\x02 \x01(\x04\"\x10\n\x0eUploadResponse\"\xa0\x01\n\x0cSignalRecord\x12\x0f\n\x07\x64\x61ta_id\x18\x01 \x01(\t\x12\x13\n\x0b\x65ndpoint_id\x18\x02 \x01(\t\x12\x38\n\rworker_signal\x18\x04 \x01(\x0b\x32!.graphsignal.signals.WorkerSignal\x12\x1a\n\x12\x64\x61ta_retention_sec\x18\x05 \x01(\x04\x12\x14\n\x0ctime_skew_ms\x18\x06 \x01(\x12\"\xb8\x07\n\x0cWorkerSignal\x12\x11\n\tworker_id\x18\x1c \x01(\t\x12\x11\n\tsignal_id\x18\x30 \x01(\t\x12\x10\n\x08\x65ndpoint\x18\x01 \x01(\t\x12\x10\n\x08start_us\x18\x05 \x01(\x04\x12\x0e\n\x06\x65nd_us\x18\x06 \x01(\x04\x12\x34\n\x0bsignal_type\x18\x31 \x01(\x0e\x32\x1f.graphsignal.signals.SignalType\x12&\n\x04tags\x18- \x03(\x0b\x32\x18.graphsignal.signals.Tag\x12\x36\n\x0cspan_metrics\x18\x33 \x01(\x0b\x32 .graphsignal.signals.SpanMetrics\x12\x35\n\x0c\x64\x61ta_metrics\x18\x34 \x03(\x0b\x32\x1f.graphsignal.signals.DataMetric\x12\x32\n\nexceptions\x18/ \x03(\x0b\x32\x1e.graphsignal.signals.Exception\x12\x36\n\nframeworks\x18$ \x03(\x0b\x32\".graphsignal.signals.FrameworkInfo\x12.\n\x08op_stats\x18\x0b \x03(\x0b\x32\x1c.graphsignal.signals.OpStats\x12\x36\n\x0ckernel_stats\x18\x0c \x03(\x0b\x32 .graphsignal.signals.KernelStats\x12\x12\n\ntrace_data\x18! \x01(\x0c\x12\x1a\n\x12is_trace_available\x18\" \x01(\x08\x12\x32\n\ndata_stats\x18\x32 \x03(\x0b\x32\x1e.graphsignal.signals.DataStats\x12\x32\n\nmodel_info\x18& \x01(\x0b\x32\x1e.graphsignal.signals.ModelInfo\x12\x36\n\x0c\x64\x65vice_usage\x18\x12 \x03(\x0b\x32 .graphsignal.signals.DeviceUsage\x12\x32\n\nnode_usage\x18\x19 \x01(\x0b\x32\x1e.graphsignal.signals.NodeUsage\x12\x38\n\rprocess_usage\x18\x1a \x01(\x0b\x32!.graphsignal.signals.ProcessUsage\x12\x32\n\nagent_info\x18# \x01(\x0b\x32\x1e.graphsignal.signals.AgentInfo\x12;\n\x0fprofiler_errors\x18\x07 \x03(\x0b\x32\".graphsignal.signals.ProfilerError\"!\n\x03Tag\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\xc8\x03\n\x06Metric\x12\x34\n\x04type\x18\x01 \x01(\x0e\x32&.graphsignal.signals.Metric.MetricType\x12\x0f\n\x05gauge\x18\x02 \x01(\x01H\x00\x12:\n\treservoir\x18\x03 \x01(\x0b\x32%.graphsignal.signals.Metric.ReservoirH\x00\x12\x36\n\x07\x63ounter\x18\x04 \x01(\x0b\x32#.graphsignal.signals.Metric.CounterH\x00\x1a|\n\x07\x43ounter\x12\x41\n\x07\x62uckets\x18\x01 \x03(\x0b\x32\x30.graphsignal.signals.Metric.Counter.BucketsEntry\x1a.\n\x0c\x42ucketsEntry\x12\x0b\n\x03key\x18\x01 \x01(\x04\x12\r\n\x05value\x18\x02 \x01(\x04:\x02\x38\x01\x1a\x1b\n\tReservoir\x12\x0e\n\x06values\x18\x01 \x03(\x01\"^\n\nMetricType\x12\x14\n\x10UNDEFINED_METRIC\x10\x00\x12\x10\n\x0cGAUGE_METRIC\x10\x01\x12\x14\n\x10RESERVOIR_METRIC\x10\x02\x12\x12\n\x0e\x43OUNTER_METRIC\x10\x03\x42\x08\n\x06values\"\xa5\x01\n\x0bSpanMetrics\x12/\n\nlatency_us\x18\x01 \x01(\x0b\x32\x1b.graphsignal.signals.Metric\x12/\n\ncall_count\x18\x02 \x01(\x0b\x32\x1b.graphsignal.signals.Metric\x12\x34\n\x0f\x65xception_count\x18\x03 \x01(\x0b\x32\x1b.graphsignal.signals.Metric\"a\n\nDataMetric\x12\x11\n\tdata_name\x18\x01 \x01(\t\x12\x13\n\x0bmetric_name\x18\x02 \x01(\t\x12+\n\x06metric\x18\x03 \x01(\x0b\x32\x1b.graphsignal.signals.Metric\"C\n\tException\x12\x10\n\x08\x65xc_type\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x13\n\x0bstack_trace\x18\x03 \x01(\t\"5\n\x06SemVer\x12\r\n\x05major\x18\x01 \x01(\x05\x12\r\n\x05minor\x18\x02 \x01(\x05\x12\r\n\x05patch\x18\x03 \x01(\x05\"\xb6\x02\n\rFrameworkInfo\x12>\n\x04type\x18\x01 \x01(\x0e\x32\x30.graphsignal.signals.FrameworkInfo.FrameworkType\x12,\n\x07version\x18\x02 \x01(\x0b\x32\x1b.graphsignal.signals.SemVer\"\xb6\x01\n\rFrameworkType\x12\x17\n\x13\x46RAMEWORK_UNDEFINED\x10\x00\x12\x18\n\x14TENSORFLOW_FRAMEWORK\x10\x01\x12\x15\n\x11PYTORCH_FRAMEWORK\x10\x02\x12\x13\n\x0fKERAS_FRAMEWORK\x10\x03\x12\x1f\n\x1bPYTORCH_LIGHTNING_FRAMEWORK\x10\x04\x12\x11\n\rJAX_FRAMEWORK\x10\x07\x12\x12\n\x0eONNX_FRAMEWORK\x10\t\"\xe9\x03\n\x07OpStats\x12\x34\n\x0b\x64\x65vice_type\x18\x01 \x01(\x0e\x32\x1f.graphsignal.signals.DeviceType\x12\x11\n\tdevice_id\x18\x02 \x01(\t\x12\x0f\n\x07op_type\x18\x03 \x01(\t\x12\x0f\n\x07op_name\x18\x04 \x01(\t\x12\r\n\x05\x63ount\x18\x05 \x01(\x04\x12\x1a\n\x12total_host_time_us\x18\x06 \x01(\x04\x12\x1c\n\x14total_device_time_us\x18\x07 \x01(\x04\x12\x19\n\x11self_host_time_us\x18\x08 \x01(\x04\x12\x1b\n\x13self_device_time_us\x18\t \x01(\x04\x12\x19\n\x11total_host_memory\x18\n \x01(\x04\x12\x1b\n\x13total_device_memory\x18\x0b \x01(\x04\x12\x18\n\x10self_host_memory\x18\x0c \x01(\x04\x12\x1a\n\x12self_device_memory\x18\r \x01(\x04\x12\x1d\n\x15self_host_memory_rate\x18\x0e \x01(\x04\x12\x1f\n\x17self_device_memory_rate\x18\x0f \x01(\x04\x12\r\n\x05\x66lops\x18\x10 \x01(\x04\x12\x15\n\rflops_per_sec\x18\x12 \x01(\x01\x12\x1e\n\x16tensorcore_utilization\x18\x11 \x01(\x01\"\xbd\x01\n\x0bKernelStats\x12\x34\n\x0b\x64\x65vice_type\x18\x01 \x01(\x0e\x32\x1f.graphsignal.signals.DeviceType\x12\x11\n\tdevice_id\x18\x02 \x01(\t\x12\x0f\n\x07op_name\x18\x03 \x01(\t\x12\x13\n\x0bkernel_name\x18\x04 \x01(\t\x12\r\n\x05\x63ount\x18\x05 \x01(\x04\x12\x13\n\x0b\x64uration_ns\x18\x06 \x01(\x04\x12\x1b\n\x13is_using_tensorcore\x18\x07 \x01(\x08\"\xa9\x01\n\tDataStats\x12\x11\n\tdata_name\x18\x01 \x01(\t\x12\x11\n\tdata_type\x18\x02 \x01(\t\x12\r\n\x05shape\x18\x04 \x03(\x04\x12=\n\x0b\x64\x61ta_counts\x18\x05 \x03(\x0b\x32(.graphsignal.signals.DataStats.DataCount\x1a(\n\tDataCount\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x63ount\x18\x02 \x01(\x04\"\xb1\x01\n\tModelInfo\x12@\n\x0cmodel_format\x18\x01 \x01(\x0e\x32*.graphsignal.signals.ModelInfo.ModelFormat\x12\x18\n\x10model_size_bytes\x18\x02 \x01(\x04\"H\n\x0bModelFormat\x12\x14\n\x10\x46ORMAT_UNDEFINED\x10\x00\x12\x12\n\x0ePYTORCH_FORMAT\x10\x01\x12\x0f\n\x0bONNX_FORMAT\x10\x02\"\xb3\x01\n\tNodeUsage\x12\x10\n\x08hostname\x18\x01 \x01(\t\x12\x12\n\nip_address\x18\x07 \x01(\t\x12\x10\n\x08mem_used\x18\t \x01(\x04\x12\x11\n\tmem_total\x18\n \x01(\x04\x12\x10\n\x08platform\x18\x02 \x01(\t\x12\x0f\n\x07machine\x18\x03 \x01(\t\x12\x0f\n\x07os_name\x18\x04 \x01(\t\x12\x12\n\nos_version\x18\x05 \x01(\t\x12\x13\n\x0bnum_devices\x18\x06 \x01(\x05\"\xce\x02\n\x0cProcessUsage\x12\x12\n\nprocess_id\x18\x05 \x01(\t\x12\x10\n\x08start_ms\x18\x0e \x01(\x04\x12\x10\n\x08\x63pu_name\x18\x10 \x01(\t\x12\x19\n\x11\x63pu_usage_percent\x18\x01 \x01(\x01\x12\x0f\n\x07max_rss\x18\x02 \x01(\x04\x12\x13\n\x0b\x63urrent_rss\x18\x03 \x01(\x04\x12\x0f\n\x07vm_size\x18\x04 \x01(\x04\x12:\n\x07runtime\x18\x06 \x01(\x0e\x32).graphsignal.signals.ProcessUsage.Runtime\x12\x34\n\x0fruntime_version\x18\x07 \x01(\x0b\x32\x1b.graphsignal.signals.SemVer\x12\x14\n\x0cruntime_impl\x18\x08 \x01(\t\",\n\x07Runtime\x12\x15\n\x11RUNTIME_UNDEFINED\x10\x00\x12\n\n\x06PYTHON\x10\x01\"\xcd\x03\n\x0b\x44\x65viceUsage\x12\x34\n\x0b\x64\x65vice_type\x18\x0b \x01(\x0e\x32\x1f.graphsignal.signals.DeviceType\x12\x11\n\tdevice_id\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65vice_name\x18\x02 \x01(\t\x12\x14\n\x0c\x61rchitecture\x18\x11 \x01(\t\x12\x37\n\x12\x63ompute_capability\x18\x0f \x01(\x0b\x32\x1b.graphsignal.signals.SemVer\x12\x11\n\tmem_total\x18\x03 \x01(\x04\x12\x10\n\x08mem_used\x18\x04 \x01(\x04\x12\x10\n\x08mem_free\x18\x05 \x01(\x04\x12\x1f\n\x17gpu_utilization_percent\x18\x06 \x01(\x01\x12\x1a\n\x12mem_access_percent\x18\x10 \x01(\x01\x12\x1a\n\x12pcie_throughput_tx\x18\r \x01(\x01\x12\x1a\n\x12pcie_throughput_rx\x18\x0e \x01(\x01\x12\x12\n\ngpu_temp_c\x18\x08 \x01(\x01\x12\x15\n\rpower_usage_w\x18\t \x01(\x01\x12\x19\n\x11\x66\x61n_speed_percent\x18\n \x01(\x01\x12\x1f\n\x17mxu_utilization_percent\x18\x0c \x01(\x01\"\x9c\x03\n\tAgentInfo\x12L\n\x17operation_profiler_type\x18\x01 \x01(\x0e\x32+.graphsignal.signals.AgentInfo.ProfilerType\x12L\n\x17\x66ramework_profiler_type\x18\x02 \x01(\x0e\x32+.graphsignal.signals.AgentInfo.ProfilerType\x12,\n\x07version\x18\x03 \x01(\x0b\x32\x1b.graphsignal.signals.SemVer\"\xc4\x01\n\x0cProfilerType\x12\x16\n\x12PROFILER_UNDEFINED\x10\x00\x12\x14\n\x10GENERIC_PROFILER\x10\x01\x12\x17\n\x13TENSORFLOW_PROFILER\x10\x02\x12\x14\n\x10PYTORCH_PROFILER\x10\x03\x12\x12\n\x0eKERAS_PROFILER\x10\x04\x12\x1e\n\x1aPYTORCH_LIGHTNING_PROFILER\x10\x06\x12\x10\n\x0cJAX_PROFILER\x10\x08\x12\x11\n\rONNX_PROFILER\x10\n\"5\n\rProfilerError\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x13\n\x0bstack_trace\x18\x02 \x01(\t*]\n\nSignalType\x12\x12\n\x0eUNKNOWN_SIGNAL\x10\x00\x12\x11\n\rSAMPLE_SIGNAL\x10\x01\x12\x14\n\x10\x45XCEPTION_SIGNAL\x10\x02\x12\x12\n\x0ePROFILE_SIGNAL\x10\x03*B\n\nDeviceType\x12\x19\n\x15\x44\x45VICE_TYPE_UNDEFINED\x10\x00\x12\x07\n\x03\x43PU\x10\x01\x12\x07\n\x03GPU\x10\x02\x12\x07\n\x03TPU\x10\x03\x62\x06proto3')

_SIGNALTYPE = DESCRIPTOR.enum_types_by_name['SignalType']
SignalType = enum_type_wrapper.EnumTypeWrapper(_SIGNALTYPE)
_DEVICETYPE = DESCRIPTOR.enum_types_by_name['DeviceType']
DeviceType = enum_type_wrapper.EnumTypeWrapper(_DEVICETYPE)
UNKNOWN_SIGNAL = 0
SAMPLE_SIGNAL = 1
EXCEPTION_SIGNAL = 2
PROFILE_SIGNAL = 3
DEVICE_TYPE_UNDEFINED = 0
CPU = 1
GPU = 2
TPU = 3


_UPLOADREQUEST = DESCRIPTOR.message_types_by_name['UploadRequest']
_UPLOADRESPONSE = DESCRIPTOR.message_types_by_name['UploadResponse']
_SIGNALRECORD = DESCRIPTOR.message_types_by_name['SignalRecord']
_WORKERSIGNAL = DESCRIPTOR.message_types_by_name['WorkerSignal']
_TAG = DESCRIPTOR.message_types_by_name['Tag']
_METRIC = DESCRIPTOR.message_types_by_name['Metric']
_METRIC_COUNTER = _METRIC.nested_types_by_name['Counter']
_METRIC_COUNTER_BUCKETSENTRY = _METRIC_COUNTER.nested_types_by_name['BucketsEntry']
_METRIC_RESERVOIR = _METRIC.nested_types_by_name['Reservoir']
_SPANMETRICS = DESCRIPTOR.message_types_by_name['SpanMetrics']
_DATAMETRIC = DESCRIPTOR.message_types_by_name['DataMetric']
_EXCEPTION = DESCRIPTOR.message_types_by_name['Exception']
_SEMVER = DESCRIPTOR.message_types_by_name['SemVer']
_FRAMEWORKINFO = DESCRIPTOR.message_types_by_name['FrameworkInfo']
_OPSTATS = DESCRIPTOR.message_types_by_name['OpStats']
_KERNELSTATS = DESCRIPTOR.message_types_by_name['KernelStats']
_DATASTATS = DESCRIPTOR.message_types_by_name['DataStats']
_DATASTATS_DATACOUNT = _DATASTATS.nested_types_by_name['DataCount']
_MODELINFO = DESCRIPTOR.message_types_by_name['ModelInfo']
_NODEUSAGE = DESCRIPTOR.message_types_by_name['NodeUsage']
_PROCESSUSAGE = DESCRIPTOR.message_types_by_name['ProcessUsage']
_DEVICEUSAGE = DESCRIPTOR.message_types_by_name['DeviceUsage']
_AGENTINFO = DESCRIPTOR.message_types_by_name['AgentInfo']
_PROFILERERROR = DESCRIPTOR.message_types_by_name['ProfilerError']
_METRIC_METRICTYPE = _METRIC.enum_types_by_name['MetricType']
_FRAMEWORKINFO_FRAMEWORKTYPE = _FRAMEWORKINFO.enum_types_by_name['FrameworkType']
_MODELINFO_MODELFORMAT = _MODELINFO.enum_types_by_name['ModelFormat']
_PROCESSUSAGE_RUNTIME = _PROCESSUSAGE.enum_types_by_name['Runtime']
_AGENTINFO_PROFILERTYPE = _AGENTINFO.enum_types_by_name['ProfilerType']
UploadRequest = _reflection.GeneratedProtocolMessageType('UploadRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPLOADREQUEST,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.UploadRequest)
  })
_sym_db.RegisterMessage(UploadRequest)

UploadResponse = _reflection.GeneratedProtocolMessageType('UploadResponse', (_message.Message,), {
  'DESCRIPTOR' : _UPLOADRESPONSE,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.UploadResponse)
  })
_sym_db.RegisterMessage(UploadResponse)

SignalRecord = _reflection.GeneratedProtocolMessageType('SignalRecord', (_message.Message,), {
  'DESCRIPTOR' : _SIGNALRECORD,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.SignalRecord)
  })
_sym_db.RegisterMessage(SignalRecord)

WorkerSignal = _reflection.GeneratedProtocolMessageType('WorkerSignal', (_message.Message,), {
  'DESCRIPTOR' : _WORKERSIGNAL,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.WorkerSignal)
  })
_sym_db.RegisterMessage(WorkerSignal)

Tag = _reflection.GeneratedProtocolMessageType('Tag', (_message.Message,), {
  'DESCRIPTOR' : _TAG,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.Tag)
  })
_sym_db.RegisterMessage(Tag)

Metric = _reflection.GeneratedProtocolMessageType('Metric', (_message.Message,), {

  'Counter' : _reflection.GeneratedProtocolMessageType('Counter', (_message.Message,), {

    'BucketsEntry' : _reflection.GeneratedProtocolMessageType('BucketsEntry', (_message.Message,), {
      'DESCRIPTOR' : _METRIC_COUNTER_BUCKETSENTRY,
      '__module__' : 'signals_pb2'
      # @@protoc_insertion_point(class_scope:graphsignal.signals.Metric.Counter.BucketsEntry)
      })
    ,
    'DESCRIPTOR' : _METRIC_COUNTER,
    '__module__' : 'signals_pb2'
    # @@protoc_insertion_point(class_scope:graphsignal.signals.Metric.Counter)
    })
  ,

  'Reservoir' : _reflection.GeneratedProtocolMessageType('Reservoir', (_message.Message,), {
    'DESCRIPTOR' : _METRIC_RESERVOIR,
    '__module__' : 'signals_pb2'
    # @@protoc_insertion_point(class_scope:graphsignal.signals.Metric.Reservoir)
    })
  ,
  'DESCRIPTOR' : _METRIC,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.Metric)
  })
_sym_db.RegisterMessage(Metric)
_sym_db.RegisterMessage(Metric.Counter)
_sym_db.RegisterMessage(Metric.Counter.BucketsEntry)
_sym_db.RegisterMessage(Metric.Reservoir)

SpanMetrics = _reflection.GeneratedProtocolMessageType('SpanMetrics', (_message.Message,), {
  'DESCRIPTOR' : _SPANMETRICS,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.SpanMetrics)
  })
_sym_db.RegisterMessage(SpanMetrics)

DataMetric = _reflection.GeneratedProtocolMessageType('DataMetric', (_message.Message,), {
  'DESCRIPTOR' : _DATAMETRIC,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.DataMetric)
  })
_sym_db.RegisterMessage(DataMetric)

Exception = _reflection.GeneratedProtocolMessageType('Exception', (_message.Message,), {
  'DESCRIPTOR' : _EXCEPTION,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.Exception)
  })
_sym_db.RegisterMessage(Exception)

SemVer = _reflection.GeneratedProtocolMessageType('SemVer', (_message.Message,), {
  'DESCRIPTOR' : _SEMVER,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.SemVer)
  })
_sym_db.RegisterMessage(SemVer)

FrameworkInfo = _reflection.GeneratedProtocolMessageType('FrameworkInfo', (_message.Message,), {
  'DESCRIPTOR' : _FRAMEWORKINFO,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.FrameworkInfo)
  })
_sym_db.RegisterMessage(FrameworkInfo)

OpStats = _reflection.GeneratedProtocolMessageType('OpStats', (_message.Message,), {
  'DESCRIPTOR' : _OPSTATS,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.OpStats)
  })
_sym_db.RegisterMessage(OpStats)

KernelStats = _reflection.GeneratedProtocolMessageType('KernelStats', (_message.Message,), {
  'DESCRIPTOR' : _KERNELSTATS,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.KernelStats)
  })
_sym_db.RegisterMessage(KernelStats)

DataStats = _reflection.GeneratedProtocolMessageType('DataStats', (_message.Message,), {

  'DataCount' : _reflection.GeneratedProtocolMessageType('DataCount', (_message.Message,), {
    'DESCRIPTOR' : _DATASTATS_DATACOUNT,
    '__module__' : 'signals_pb2'
    # @@protoc_insertion_point(class_scope:graphsignal.signals.DataStats.DataCount)
    })
  ,
  'DESCRIPTOR' : _DATASTATS,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.DataStats)
  })
_sym_db.RegisterMessage(DataStats)
_sym_db.RegisterMessage(DataStats.DataCount)

ModelInfo = _reflection.GeneratedProtocolMessageType('ModelInfo', (_message.Message,), {
  'DESCRIPTOR' : _MODELINFO,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.ModelInfo)
  })
_sym_db.RegisterMessage(ModelInfo)

NodeUsage = _reflection.GeneratedProtocolMessageType('NodeUsage', (_message.Message,), {
  'DESCRIPTOR' : _NODEUSAGE,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.NodeUsage)
  })
_sym_db.RegisterMessage(NodeUsage)

ProcessUsage = _reflection.GeneratedProtocolMessageType('ProcessUsage', (_message.Message,), {
  'DESCRIPTOR' : _PROCESSUSAGE,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.ProcessUsage)
  })
_sym_db.RegisterMessage(ProcessUsage)

DeviceUsage = _reflection.GeneratedProtocolMessageType('DeviceUsage', (_message.Message,), {
  'DESCRIPTOR' : _DEVICEUSAGE,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.DeviceUsage)
  })
_sym_db.RegisterMessage(DeviceUsage)

AgentInfo = _reflection.GeneratedProtocolMessageType('AgentInfo', (_message.Message,), {
  'DESCRIPTOR' : _AGENTINFO,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.AgentInfo)
  })
_sym_db.RegisterMessage(AgentInfo)

ProfilerError = _reflection.GeneratedProtocolMessageType('ProfilerError', (_message.Message,), {
  'DESCRIPTOR' : _PROFILERERROR,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.ProfilerError)
  })
_sym_db.RegisterMessage(ProfilerError)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _METRIC_COUNTER_BUCKETSENTRY._options = None
  _METRIC_COUNTER_BUCKETSENTRY._serialized_options = b'8\001'
  _SIGNALTYPE._serialized_start=4956
  _SIGNALTYPE._serialized_end=5049
  _DEVICETYPE._serialized_start=5051
  _DEVICETYPE._serialized_end=5117
  _UPLOADREQUEST._serialized_start=38
  _UPLOADREQUEST._serialized_end=131
  _UPLOADRESPONSE._serialized_start=133
  _UPLOADRESPONSE._serialized_end=149
  _SIGNALRECORD._serialized_start=152
  _SIGNALRECORD._serialized_end=312
  _WORKERSIGNAL._serialized_start=315
  _WORKERSIGNAL._serialized_end=1267
  _TAG._serialized_start=1269
  _TAG._serialized_end=1302
  _METRIC._serialized_start=1305
  _METRIC._serialized_end=1761
  _METRIC_COUNTER._serialized_start=1502
  _METRIC_COUNTER._serialized_end=1626
  _METRIC_COUNTER_BUCKETSENTRY._serialized_start=1580
  _METRIC_COUNTER_BUCKETSENTRY._serialized_end=1626
  _METRIC_RESERVOIR._serialized_start=1628
  _METRIC_RESERVOIR._serialized_end=1655
  _METRIC_METRICTYPE._serialized_start=1657
  _METRIC_METRICTYPE._serialized_end=1751
  _SPANMETRICS._serialized_start=1764
  _SPANMETRICS._serialized_end=1929
  _DATAMETRIC._serialized_start=1931
  _DATAMETRIC._serialized_end=2028
  _EXCEPTION._serialized_start=2030
  _EXCEPTION._serialized_end=2097
  _SEMVER._serialized_start=2099
  _SEMVER._serialized_end=2152
  _FRAMEWORKINFO._serialized_start=2155
  _FRAMEWORKINFO._serialized_end=2465
  _FRAMEWORKINFO_FRAMEWORKTYPE._serialized_start=2283
  _FRAMEWORKINFO_FRAMEWORKTYPE._serialized_end=2465
  _OPSTATS._serialized_start=2468
  _OPSTATS._serialized_end=2957
  _KERNELSTATS._serialized_start=2960
  _KERNELSTATS._serialized_end=3149
  _DATASTATS._serialized_start=3152
  _DATASTATS._serialized_end=3321
  _DATASTATS_DATACOUNT._serialized_start=3281
  _DATASTATS_DATACOUNT._serialized_end=3321
  _MODELINFO._serialized_start=3324
  _MODELINFO._serialized_end=3501
  _MODELINFO_MODELFORMAT._serialized_start=3429
  _MODELINFO_MODELFORMAT._serialized_end=3501
  _NODEUSAGE._serialized_start=3504
  _NODEUSAGE._serialized_end=3683
  _PROCESSUSAGE._serialized_start=3686
  _PROCESSUSAGE._serialized_end=4020
  _PROCESSUSAGE_RUNTIME._serialized_start=3976
  _PROCESSUSAGE_RUNTIME._serialized_end=4020
  _DEVICEUSAGE._serialized_start=4023
  _DEVICEUSAGE._serialized_end=4484
  _AGENTINFO._serialized_start=4487
  _AGENTINFO._serialized_end=4899
  _AGENTINFO_PROFILERTYPE._serialized_start=4703
  _AGENTINFO_PROFILERTYPE._serialized_end=4899
  _PROFILERERROR._serialized_start=4901
  _PROFILERERROR._serialized_end=4954
# @@protoc_insertion_point(module_scope)



from typing import Optional
import logging
import os
import sys
import time
import gzip
import jax

import graphsignal
from graphsignal.proto_utils import parse_semver, compare_semver
from graphsignal.proto import signals_pb2
from graphsignal.profilers.operation_profiler import OperationProfiler
from graphsignal.profilers.profiler_utils import create_log_dir, remove_log_dir, convert_tensorflow_profile

logger = logging.getLogger('graphsignal')

class JaxProfiler(OperationProfiler):
    def __init__(self):
        self._log_dir = None
        self._jax_version = None

    def read_info(self, signal):
        if not self._jax_version:
            self._jax_version = signals_pb2.SemVer()
            parse_semver(self._jax_version, jax.__version__)

        # Framework info
        framework = signal.frameworks.add()
        framework.type = signals_pb2.FrameworkInfo.FrameworkType.JAX_FRAMEWORK
        framework.version.CopyFrom(self._jax_version)

    def start(self, signal):
        logger.debug('Activating JAX profiler')

        # Profiler info
        signal.agent_info.operation_profiler_type = signals_pb2.AgentInfo.ProfilerType.JAX_PROFILER

        try:
            self._log_dir = create_log_dir()

            jax.profiler.start_trace(self._log_dir)
        except Exception as e:
            remove_log_dir(self._log_dir)
            raise e

    def stop(self, signal):
        logger.debug('Deactivating JAX profiler')

        try:
            jax.profiler.stop_trace()

            convert_tensorflow_profile(self._log_dir, signal)
        finally:
            remove_log_dir(self._log_dir)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: graphsignal/profilers/tensorflow_proto/diagnostics.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n8graphsignal/profilers/tensorflow_proto/diagnostics.proto\x12\x13tensorflow.profiler\"=\n\x0b\x44iagnostics\x12\x0c\n\x04info\x18\x01 \x03(\t\x12\x10\n\x08warnings\x18\x02 \x03(\t\x12\x0e\n\x06\x65rrors\x18\x03 \x03(\tb\x06proto3')



_DIAGNOSTICS = DESCRIPTOR.message_types_by_name['Diagnostics']
Diagnostics = _reflection.GeneratedProtocolMessageType('Diagnostics', (_message.Message,), {
  'DESCRIPTOR' : _DIAGNOSTICS,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.diagnostics_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.Diagnostics)
  })
_sym_db.RegisterMessage(Diagnostics)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _DIAGNOSTICS._serialized_start=81
  _DIAGNOSTICS._serialized_end=142
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: graphsignal/profilers/tensorflow_proto/kernel_stats.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n9graphsignal/profilers/tensorflow_proto/kernel_stats.proto\x12\x13tensorflow.profiler\"\xeb\x02\n\x0cKernelReport\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1c\n\x14registers_per_thread\x18\x02 \x01(\r\x12\x1a\n\x12static_shmem_bytes\x18\x03 \x01(\r\x12\x1b\n\x13\x64ynamic_shmem_bytes\x18\x04 \x01(\r\x12\x11\n\tblock_dim\x18\x05 \x03(\r\x12\x10\n\x08grid_dim\x18\x06 \x03(\r\x12\x19\n\x11total_duration_ns\x18\x07 \x01(\x04\x12\x17\n\x0fmin_duration_ns\x18\x08 \x01(\x04\x12\x17\n\x0fmax_duration_ns\x18\t \x01(\x04\x12#\n\x1bis_kernel_using_tensor_core\x18\n \x01(\x08\x12\"\n\x1ais_op_tensor_core_eligible\x18\x0b \x01(\x08\x12\x0f\n\x07op_name\x18\x0c \x01(\t\x12\x13\n\x0boccurrences\x18\r \x01(\r\x12\x15\n\roccupancy_pct\x18\x0e \x01(\x02\"C\n\rKernelStatsDb\x12\x32\n\x07reports\x18\x01 \x03(\x0b\x32!.tensorflow.profiler.KernelReportb\x06proto3')



_KERNELREPORT = DESCRIPTOR.message_types_by_name['KernelReport']
_KERNELSTATSDB = DESCRIPTOR.message_types_by_name['KernelStatsDb']
KernelReport = _reflection.GeneratedProtocolMessageType('KernelReport', (_message.Message,), {
  'DESCRIPTOR' : _KERNELREPORT,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.kernel_stats_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.KernelReport)
  })
_sym_db.RegisterMessage(KernelReport)

KernelStatsDb = _reflection.GeneratedProtocolMessageType('KernelStatsDb', (_message.Message,), {
  'DESCRIPTOR' : _KERNELSTATSDB,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.kernel_stats_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.KernelStatsDb)
  })
_sym_db.RegisterMessage(KernelStatsDb)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _KERNELREPORT._serialized_start=83
  _KERNELREPORT._serialized_end=446
  _KERNELSTATSDB._serialized_start=448
  _KERNELSTATSDB._serialized_end=515
# @@protoc_insertion_point(module_scope)



import unittest
import logging
import sys
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')


class HuggingFaceGeneratorTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_pipeline(self):
        from transformers import pipeline
        from graphsignal.profilers.pytorch import PyTorchProfiler

        pipe = pipeline(task="text-generation", model='distilgpt2')

        profiler = PyTorchProfiler()
        signal = signals_pb2.WorkerSignal()
        profiler.start(signal)
        output = pipe('some text')
        profiler.stop(signal)

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(signal))

        test_op_stats = None
        for op_stats in signal.op_stats:
            if op_stats.op_name == 'aten::mm':
                test_op_stats = op_stats
                break
        self.assertIsNotNone(test_op_stats)
        self.assertTrue(test_op_stats.count >= 1)
        self.assertTrue(test_op_stats.total_host_time_us >= 1)
        self.assertTrue(test_op_stats.self_host_time_us >= 1)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: graphsignal/profilers/tensorflow_proto/overview_page.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from graphsignal.profilers.tensorflow_proto import diagnostics_pb2 as graphsignal_dot_profilers_dot_tensorflow__proto_dot_diagnostics__pb2
from graphsignal.profilers.tensorflow_proto import input_pipeline_pb2 as graphsignal_dot_profilers_dot_tensorflow__proto_dot_input__pipeline__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n:graphsignal/profilers/tensorflow_proto/overview_page.proto\x12\x13tensorflow.profiler\x1a\x19google/protobuf/any.proto\x1a\x38graphsignal/profilers/tensorflow_proto/diagnostics.proto\x1a;graphsignal/profilers/tensorflow_proto/input_pipeline.proto\"\xc2\x01\n\x0cOverviewTfOp\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x63\x61tegory\x18\x02 \x01(\t\x12\x1a\n\x12self_time_fraction\x18\x03 \x01(\x01\x12 \n\x18\x63umulative_time_fraction\x18\x04 \x01(\x01\x12\x11\n\tflop_rate\x18\x05 \x01(\x01\x12!\n\x19is_op_tensorcore_eligible\x18\x06 \x01(\x08\x12\x1e\n\x16is_op_using_tensorcore\x18\x07 \x01(\x08\"\xf5\x04\n\x14OverviewPageAnalysis\x12\x1f\n\x17mxu_utilization_percent\x18\x01 \x01(\x01\x12 \n\x18\x64\x65vice_idle_time_percent\x18\x02 \x01(\x01\x12\x1e\n\x16host_idle_time_percent\x18\x03 \x01(\x01\x12\x39\n\x0etop_device_ops\x18\x04 \x03(\x0b\x32!.tensorflow.profiler.OverviewTfOp\x12\x13\n\x0bremark_text\x18\x05 \x01(\t\x12\x14\n\x0cremark_color\x18\x06 \x01(\t\x12:\n2flop_rate_utilization_relative_to_roofline_percent\x18\x07 \x01(\x01\x12:\n2memory_bw_utilization_relative_to_hw_limit_percent\x18\x08 \x01(\x01\x12$\n\x1c\x64\x65vice_compute_16bit_percent\x18\t \x01(\x01\x12$\n\x1c\x64\x65vice_compute_32bit_percent\x18\n \x01(\x01\x12\x1a\n\x12host_tf_op_percent\x18\x0b \x01(\x01\x12\x1c\n\x14\x64\x65vice_tf_op_percent\x18\x0c \x01(\x01\x12\x18\n\x10host_trace_level\x18\r \x01(\r\x12\"\n\x1ahost_op_time_eager_percent\x18\x0e \x01(\x01\x12$\n\x1c\x64\x65vice_op_time_eager_percent\x18\x0f \x01(\x01\x12\x32\n*device_op_time_outside_compilation_percent\x18\x10 \x01(\x01\"\x1f\n\x0fOverviewPageTip\x12\x0c\n\x04link\x18\x01 \x01(\t\"\xff\x01\n\x15GenericRecommendation\x12 \n\x18kernel_launch_bottleneck\x18\x01 \x01(\t\x12\x1f\n\x17kernel_launch_statement\x18\x02 \x01(\t\x12\x1c\n\x14\x61ll_other_bottleneck\x18\x03 \x01(\t\x12\x1b\n\x13\x61ll_other_statement\x18\x04 \x01(\t\x12\x1b\n\x13precision_statement\x18\x05 \x01(\t\x12%\n\x1d\x64\x65vice_collectives_bottleneck\x18\x06 \x01(\t\x12$\n\x1c\x64\x65vice_collectives_statement\x18\x07 \x01(\t\"\xdf\x04\n\x1aOverviewPageRecommendation\x12\x12\n\nbottleneck\x18\x01 \x01(\t\x12\x11\n\tstatement\x18\x02 \x01(\t\x12\x38\n\ninput_tips\x18\x0b \x03(\x0b\x32$.tensorflow.profiler.OverviewPageTip\x12\x18\n\x10output_statement\x18\t \x01(\t\x12\x1c\n\x14\x65\x61ger_statement_html\x18\x0c \x01(\t\x12*\n\"outside_compilation_statement_html\x18\r \x01(\t\x12\"\n\x1atf_function_statement_html\x18\n \x01(\t\x12\x37\n\thost_tips\x18\x03 \x03(\x0b\x32$.tensorflow.profiler.OverviewPageTip\x12\x39\n\x0b\x64\x65vice_tips\x18\x04 \x03(\x0b\x32$.tensorflow.profiler.OverviewPageTip\x12@\n\x12\x64ocumentation_tips\x18\x05 \x03(\x0b\x32$.tensorflow.profiler.OverviewPageTip\x12,\n\x0erecommendation\x18\x06 \x01(\x0b\x32\x14.google.protobuf.Any\x12\x36\n\x08\x66\x61q_tips\x18\x07 \x03(\x0b\x32$.tensorflow.profiler.OverviewPageTip\x12<\n\x0einference_tips\x18\x08 \x03(\x0b\x32$.tensorflow.profiler.OverviewPageTip\"\x80\x01\n\"OverviewPageHostIndependentJobInfo\x12\x13\n\x0b\x63hange_list\x18\x01 \x01(\x03\x12\x12\n\nbuild_time\x18\x02 \x01(\x03\x12\x14\n\x0c\x62uild_target\x18\x03 \x01(\t\x12\x1b\n\x13profile_duration_ms\x18\x04 \x01(\r\"\x8b\x01\n OverviewPageHostDependentJobInfo\x12\x0f\n\x07host_id\x18\x01 \x01(\t\x12\x14\n\x0c\x63ommand_line\x18\x02 \x01(\t\x12\x12\n\nstart_time\x18\x03 \x01(\x03\x12\x13\n\x0b\x62ns_address\x18\x04 \x01(\t\x12\x17\n\x0fprofile_time_ns\x18\x05 \x01(\x04\"\x80\x04\n\x1aOverviewPageRunEnvironment\x12\x12\n\nhost_count\x18\x01 \x01(\x05\x12\x12\n\ntask_count\x18\x02 \x01(\x05\x12Q\n\thostnames\x18\x03 \x03(\x0b\x32>.tensorflow.profiler.OverviewPageRunEnvironment.HostnamesEntry\x12\x13\n\x0b\x64\x65vice_type\x18\x04 \x01(\t\x12\x19\n\x11\x64\x65vice_core_count\x18\x05 \x01(\x05\x12\x1b\n\x13per_core_batch_size\x18\x06 \x01(\x05\x12Z\n\x19host_independent_job_info\x18\x07 \x01(\x0b\x32\x37.tensorflow.profiler.OverviewPageHostIndependentJobInfo\x12V\n\x17host_dependent_job_info\x18\x08 \x03(\x0b\x32\x35.tensorflow.profiler.OverviewPageHostDependentJobInfo\x12\x15\n\rreplica_count\x18\t \x01(\x05\x12\x1d\n\x15num_cores_per_replica\x18\n \x01(\x05\x1a\x30\n\x0eHostnamesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x08:\x02\x38\x01\"\xf1\x02\n\x0cOverviewPage\x12H\n\x0frun_environment\x18\x06 \x01(\x0b\x32/.tensorflow.profiler.OverviewPageRunEnvironment\x12H\n\x0einput_analysis\x18\x02 \x01(\x0b\x32\x30.tensorflow.profiler.InputPipelineAnalysisResult\x12;\n\x08\x61nalysis\x18\x03 \x01(\x0b\x32).tensorflow.profiler.OverviewPageAnalysis\x12G\n\x0erecommendation\x18\x04 \x01(\x0b\x32/.tensorflow.profiler.OverviewPageRecommendation\x12\x35\n\x0b\x64iagnostics\x18\x08 \x01(\x0b\x32 .tensorflow.profiler.DiagnosticsJ\x04\x08\x01\x10\x02J\x04\x08\x05\x10\x06J\x04\x08\x07\x10\x08\x62\x06proto3')



_OVERVIEWTFOP = DESCRIPTOR.message_types_by_name['OverviewTfOp']
_OVERVIEWPAGEANALYSIS = DESCRIPTOR.message_types_by_name['OverviewPageAnalysis']
_OVERVIEWPAGETIP = DESCRIPTOR.message_types_by_name['OverviewPageTip']
_GENERICRECOMMENDATION = DESCRIPTOR.message_types_by_name['GenericRecommendation']
_OVERVIEWPAGERECOMMENDATION = DESCRIPTOR.message_types_by_name['OverviewPageRecommendation']
_OVERVIEWPAGEHOSTINDEPENDENTJOBINFO = DESCRIPTOR.message_types_by_name['OverviewPageHostIndependentJobInfo']
_OVERVIEWPAGEHOSTDEPENDENTJOBINFO = DESCRIPTOR.message_types_by_name['OverviewPageHostDependentJobInfo']
_OVERVIEWPAGERUNENVIRONMENT = DESCRIPTOR.message_types_by_name['OverviewPageRunEnvironment']
_OVERVIEWPAGERUNENVIRONMENT_HOSTNAMESENTRY = _OVERVIEWPAGERUNENVIRONMENT.nested_types_by_name['HostnamesEntry']
_OVERVIEWPAGE = DESCRIPTOR.message_types_by_name['OverviewPage']
OverviewTfOp = _reflection.GeneratedProtocolMessageType('OverviewTfOp', (_message.Message,), {
  'DESCRIPTOR' : _OVERVIEWTFOP,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.overview_page_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.OverviewTfOp)
  })
_sym_db.RegisterMessage(OverviewTfOp)

OverviewPageAnalysis = _reflection.GeneratedProtocolMessageType('OverviewPageAnalysis', (_message.Message,), {
  'DESCRIPTOR' : _OVERVIEWPAGEANALYSIS,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.overview_page_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.OverviewPageAnalysis)
  })
_sym_db.RegisterMessage(OverviewPageAnalysis)

OverviewPageTip = _reflection.GeneratedProtocolMessageType('OverviewPageTip', (_message.Message,), {
  'DESCRIPTOR' : _OVERVIEWPAGETIP,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.overview_page_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.OverviewPageTip)
  })
_sym_db.RegisterMessage(OverviewPageTip)

GenericRecommendation = _reflection.GeneratedProtocolMessageType('GenericRecommendation', (_message.Message,), {
  'DESCRIPTOR' : _GENERICRECOMMENDATION,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.overview_page_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.GenericRecommendation)
  })
_sym_db.RegisterMessage(GenericRecommendation)

OverviewPageRecommendation = _reflection.GeneratedProtocolMessageType('OverviewPageRecommendation', (_message.Message,), {
  'DESCRIPTOR' : _OVERVIEWPAGERECOMMENDATION,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.overview_page_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.OverviewPageRecommendation)
  })
_sym_db.RegisterMessage(OverviewPageRecommendation)

OverviewPageHostIndependentJobInfo = _reflection.GeneratedProtocolMessageType('OverviewPageHostIndependentJobInfo', (_message.Message,), {
  'DESCRIPTOR' : _OVERVIEWPAGEHOSTINDEPENDENTJOBINFO,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.overview_page_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.OverviewPageHostIndependentJobInfo)
  })
_sym_db.RegisterMessage(OverviewPageHostIndependentJobInfo)

OverviewPageHostDependentJobInfo = _reflection.GeneratedProtocolMessageType('OverviewPageHostDependentJobInfo', (_message.Message,), {
  'DESCRIPTOR' : _OVERVIEWPAGEHOSTDEPENDENTJOBINFO,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.overview_page_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.OverviewPageHostDependentJobInfo)
  })
_sym_db.RegisterMessage(OverviewPageHostDependentJobInfo)

OverviewPageRunEnvironment = _reflection.GeneratedProtocolMessageType('OverviewPageRunEnvironment', (_message.Message,), {

  'HostnamesEntry' : _reflection.GeneratedProtocolMessageType('HostnamesEntry', (_message.Message,), {
    'DESCRIPTOR' : _OVERVIEWPAGERUNENVIRONMENT_HOSTNAMESENTRY,
    '__module__' : 'graphsignal.profilers.tensorflow_proto.overview_page_pb2'
    # @@protoc_insertion_point(class_scope:tensorflow.profiler.OverviewPageRunEnvironment.HostnamesEntry)
    })
  ,
  'DESCRIPTOR' : _OVERVIEWPAGERUNENVIRONMENT,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.overview_page_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.OverviewPageRunEnvironment)
  })
_sym_db.RegisterMessage(OverviewPageRunEnvironment)
_sym_db.RegisterMessage(OverviewPageRunEnvironment.HostnamesEntry)

OverviewPage = _reflection.GeneratedProtocolMessageType('OverviewPage', (_message.Message,), {
  'DESCRIPTOR' : _OVERVIEWPAGE,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.overview_page_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.OverviewPage)
  })
_sym_db.RegisterMessage(OverviewPage)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _OVERVIEWPAGERUNENVIRONMENT_HOSTNAMESENTRY._options = None
  _OVERVIEWPAGERUNENVIRONMENT_HOSTNAMESENTRY._serialized_options = b'8\001'
  _OVERVIEWTFOP._serialized_start=230
  _OVERVIEWTFOP._serialized_end=424
  _OVERVIEWPAGEANALYSIS._serialized_start=427
  _OVERVIEWPAGEANALYSIS._serialized_end=1056
  _OVERVIEWPAGETIP._serialized_start=1058
  _OVERVIEWPAGETIP._serialized_end=1089
  _GENERICRECOMMENDATION._serialized_start=1092
  _GENERICRECOMMENDATION._serialized_end=1347
  _OVERVIEWPAGERECOMMENDATION._serialized_start=1350
  _OVERVIEWPAGERECOMMENDATION._serialized_end=1957
  _OVERVIEWPAGEHOSTINDEPENDENTJOBINFO._serialized_start=1960
  _OVERVIEWPAGEHOSTINDEPENDENTJOBINFO._serialized_end=2088
  _OVERVIEWPAGEHOSTDEPENDENTJOBINFO._serialized_start=2091
  _OVERVIEWPAGEHOSTDEPENDENTJOBINFO._serialized_end=2230
  _OVERVIEWPAGERUNENVIRONMENT._serialized_start=2233
  _OVERVIEWPAGERUNENVIRONMENT._serialized_end=2745
  _OVERVIEWPAGERUNENVIRONMENT_HOSTNAMESENTRY._serialized_start=2697
  _OVERVIEWPAGERUNENVIRONMENT_HOSTNAMESENTRY._serialized_end=2745
  _OVERVIEWPAGE._serialized_start=2748
  _OVERVIEWPAGE._serialized_end=3117
# @@protoc_insertion_point(module_scope)



import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.trace_sampler import TraceSampler

logger = logging.getLogger('graphsignal')


class TraceSamplerTest(unittest.TestCase):
    def setUp(self):
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_ensured(self):
        sampler = TraceSampler()
        for _ in range(TraceSampler.MAX_ENSURED_SPANS):
            self.assertTrue(sampler.lock(ensure=True))
            sampler.unlock()
        sampler.unlock()

    def test_predefined(self):
        sampler = TraceSampler()
        first_span = next(iter(sampler._span_filter))
        for idx in range(TraceSampler.PREDEFINED_SPANS[-1]):
            if idx + 1 in sampler._span_filter:
                self.assertTrue(sampler.lock())
                sampler.unlock()
            else:
                self.assertFalse(sampler.lock())
                sampler.unlock()
        self.assertFalse(sampler.lock())
        sampler.unlock()

    def test_interval(self):
        sampler = TraceSampler()
        sampler._span_filter = {}
        self.assertFalse(sampler.lock())
        sampler.unlock()
        self.assertFalse(sampler._interval_mode)

        # first interval
        sampler._start_ts = time.time() - TraceSampler.MIN_INTERVAL_SEC - 10
        self.assertTrue(sampler.lock())
        sampler.unlock()
        self.assertFalse(sampler.lock())
        sampler.unlock()
        self.assertTrue(sampler._interval_mode)

        # other intervals
        sampler._start_ts = time.time() - TraceSampler.MIN_INTERVAL_SEC - 10
        sampler._last_sample_ts = time.time() - TraceSampler.MIN_INTERVAL_SEC - 10
        self.assertTrue(sampler.lock())
        sampler.unlock()
        self.assertFalse(sampler.lock())
        sampler.unlock()

        # test ensured during interval
        self.assertFalse(sampler.lock())
        for _ in range(TraceSampler.MAX_ENSURED_SPANS):
            self.assertTrue(sampler.lock(ensure=True))
            sampler.unlock()
        self.assertFalse(sampler.lock())
        sampler.unlock()

    def test_concurrent(self):
        sampler = TraceSampler()
        self.assertTrue(sampler.lock(ensure=True))
        self.assertFalse(sampler.lock(ensure=True))
        self.assertFalse(sampler.lock(ensure=True))
        sampler.unlock()
        self.assertTrue(sampler.lock(ensure=True))
        sampler.unlock()




import unittest
import logging
import sys
from unittest.mock import patch, Mock
import torch
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.profilers.pytorch import PyTorchProfiler
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')


class PyTorchProfilerTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
        graphsignal.shutdown()

    def test_read_info(self):
        profiler = PyTorchProfiler()
        signal = signals_pb2.WorkerSignal()
        profiler.read_info(signal)

        self.assertEqual(
            signal.frameworks[0].type,
            signals_pb2.FrameworkInfo.FrameworkType.PYTORCH_FRAMEWORK)

    def test_start_stop(self):
        x = torch.arange(-5, 5, 0.1).view(-1, 1)
        y = -5 * x + 0.1 * torch.randn(x.size())
        model = torch.nn.Linear(1, 1)
        if torch.cuda.is_available():
            x = x.to('cuda:0')
            y = y.to('cuda:0')
            model = model.to('cuda:0')

        profiler = PyTorchProfiler()
        signal = signals_pb2.WorkerSignal()
        profiler.start(signal)
        y1 = model(x)
        profiler.stop(signal)

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(signal))

        test_op_stats = None
        for op_stats in signal.op_stats:
            if op_stats.op_name == 'aten::addmm':
                test_op_stats = op_stats
                break
        self.assertIsNotNone(test_op_stats)
        self.assertTrue(test_op_stats.count >= 1)
        if torch.cuda.is_available():
            self.assertTrue(test_op_stats.total_device_time_us >= 1)
            self.assertTrue(test_op_stats.self_device_time_us >= 1)
        else:
            self.assertTrue(test_op_stats.total_host_time_us >= 1)
            self.assertTrue(test_op_stats.self_host_time_us >= 1)

        self.assertNotEqual(signal.trace_data, b'')



from typing import Union, Any, Optional, Dict
import logging
import sys
import time
import traceback

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.data import compute_counts, build_stats 

logger = logging.getLogger('graphsignal')


class TraceSpan:
    MAX_TAGS = 10
    MAX_DATA_OBJECTS = 10

    __slots__ = [
        '_operation_profiler',
        '_trace_sampler',
        '_metric_store',
        '_agent',
        '_endpoint',
        '_ensure_trace',
        '_tags',
        '_is_tracing',
        '_is_profiling',
        '_signal',
        '_is_stopped',
        '_start_counter',
        '_exc_info',
        '_data'
    ]

    def __init__(self, 
            endpoint,
            tags=None,
            ensure_trace=False, 
            operation_profiler=None):
        if not endpoint:
            raise ValueError('endpoint is required')
        if not isinstance(endpoint, str):
            raise ValueError('endpoint must be string')
        if len(endpoint) > 50:
            raise ValueError('endpoint is too long (>50)')
        if tags is not None:
            if not isinstance(tags, dict):
                raise ValueError('tags must be dict')
            if len(tags) > TraceSpan.MAX_TAGS:
                raise ValueError('too many tags (>{0})'.format(TraceSpan.MAX_TAGS))

        self._endpoint = endpoint
        self._tags = tags
        self._ensure_trace = ensure_trace
        self._operation_profiler = operation_profiler

        self._agent = None
        self._trace_sampler = None
        self._metric_store = None
        self._is_stopped = False
        self._is_tracing = False
        self._is_profiling = False
        self._signal = None
        self._exc_info = None
        self._data = None

        try:
            self._start()
        except Exception as ex:
            if self._is_tracing:
                self._is_stopped = True
                self._trace_sampler.unlock()
            raise ex

    def is_tracing(self):
        return self._is_tracing

    def is_profiling(self):
        return self._is_profiling

    def _start(self):
        if self._is_stopped:
            return

        self._agent = graphsignal._agent
        self._trace_sampler = self._agent.get_trace_sampler(self._endpoint)
        self._metric_store = self._agent.get_metric_store(self._endpoint)

        if self._trace_sampler.lock(ensure=self._ensure_trace):
            if logger.isEnabledFor(logging.DEBUG):
                profiling_start_overhead_counter = time.perf_counter()

            self._is_tracing = True
            self._signal = self._agent.create_signal()

            # read framework-specific info
            if self._operation_profiler:
                self._operation_profiler.read_info(self._signal)

            # start profiler
            if self._operation_profiler and self._trace_sampler.should_profile():
                try:
                    self._operation_profiler.start(self._signal)
                    self._is_profiling = True
                except Exception as exc:
                    self._add_profiler_exception(exc)
                    logger.error('Error starting profiler', exc_info=True)

            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('Profiling start took: %fs', time.perf_counter() - profiling_start_overhead_counter)

        self._start_counter = time.perf_counter()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self._exc_info = exc_info
        self.stop()
        return False

    def stop(self) -> None:
        try:
            self._stop()
        finally:
            if self._is_tracing:
                self._is_stopped = True
                self._trace_sampler.unlock()            

    def _stop(self) -> None:
        stop_counter = time.perf_counter()
        duration_us = int((stop_counter - self._start_counter) * 1e6)
        end_us = _timestamp_us()

        if self._is_stopped:
            return
        self._is_stopped = True

        # if exception, but the span is not being traced, try to start tracing
        if self._exc_info and self._exc_info[0] and not self._is_tracing:
            if self._trace_sampler.lock(ensure=True):
                self._is_tracing = True
                self._signal = self._agent.create_signal()

        # stop profiler, if profiling
        if self._is_profiling:
            try:
                self._operation_profiler.stop(self._signal)
            except Exception as exc:
                logger.error('Error stopping profiler', exc_info=True)
                self._add_profiler_exception(exc)

        # update time and counters
        if not self._is_profiling:
            # only measure time if not profiling due to profiler overhead
            self._metric_store.add_time(duration_us)
        self._metric_store.inc_call_count(1, end_us)
        if self._data is not None:
            for data_name, data in self._data.items():
                data_counts = compute_counts(data)
                for count_name, count in data_counts.items():
                    self._metric_store.inc_data_counter(
                        data_name, count_name, count, end_us)

        # update exception counter
        if self._exc_info and self._exc_info[0]:
            self._metric_store.inc_exception_count(1, end_us)

        # fill and upload profile
        if self._is_tracing:
            if logger.isEnabledFor(logging.DEBUG):
                profiling_stop_overhead_counter = time.perf_counter()

            # read usage data
            try:
                self._agent.read_usage(self._signal)
            except Exception as exc:
                logger.error('Error reading usage information', exc_info=True)
                self._add_profiler_exception(exc)

            # copy data to profile
            self._signal.endpoint = self._endpoint
            self._signal.start_us = end_us - duration_us
            self._signal.end_us = end_us
            if self._exc_info and self._exc_info[0]:
                self._signal.signal_type = signals_pb2.SignalType.EXCEPTION_SIGNAL
            elif self._is_profiling:
                self._signal.signal_type = signals_pb2.SignalType.PROFILE_SIGNAL
            else:
                self._signal.signal_type = signals_pb2.SignalType.SAMPLE_SIGNAL

            # copy tags
            if self._tags is not None:
                for key, value in self._tags.items():
                    tag = self._signal.tags.add()
                    tag.key = key[:50]
                    tag.value = str(value)[:50]

            # copy inference stats
            self._metric_store.finalize(end_us)
            if self._metric_store.latency_us:
                self._signal.span_metrics.latency_us.CopyFrom(
                    self._metric_store.latency_us)
            if self._metric_store.call_count:
                self._signal.span_metrics.call_count.CopyFrom(
                    self._metric_store.call_count)
            if self._metric_store.exception_count:
                self._signal.span_metrics.exception_count.CopyFrom(
                    self._metric_store.exception_count)
            for counter in self._metric_store.data_counters.values():
                self._signal.data_metrics.append(counter)
            self._agent.reset_metric_store(self._endpoint)

            # copy exception
            if self._exc_info and self._exc_info[0]:
                exception = self._signal.exceptions.add()
                if self._exc_info[0] and hasattr(self._exc_info[0], '__name__'):
                    exception.exc_type = str(self._exc_info[0].__name__)
                if self._exc_info[1]:
                    exception.message = str(self._exc_info[1])
                if self._exc_info[2]:
                    frames = traceback.format_tb(self._exc_info[2])
                    if len(frames) > 0:
                        exception.stack_trace = ''.join(frames)

            # copy data stats
            if self._data is not None:
                for name, data in self._data.items():
                    try:
                        data_stats_proto = build_stats(data)
                        data_stats_proto.data_name = name
                        if data_stats_proto:
                            self._signal.data_stats.append(data_stats_proto)
                    except Exception as exc:
                        logger.error('Error computing data stats', exc_info=True)
                        self._add_profiler_exception(exc)

            # queue signal for upload
            self._agent.uploader().upload_signal(self._signal)
            self._agent.tick()

            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('Profiling stop took: %fs', time.perf_counter() - profiling_stop_overhead_counter)

    def set_tag(self, key: str, value: str) -> None:
        if not key:
            raise ValueError('set_tag: key must be provided')
        if value is None:
            raise ValueError('set_tag: value must be provided')

        if self._tags is None:
            self._tags = {}

        if len(self._tags) > TraceSpan.MAX_TAGS:
            raise ValueError('set_tag: too many tags (>{0})'.format(TraceSpan.MAX_TAGS))

        self._tags[key] = value

    def set_exception(self, exc: Optional[Exception] = None, exc_info: Optional[bool] = None) -> None:
        if exc is not None and not isinstance(exc, Exception):
            raise ValueError('set_exception: exc must be instance of Exception')

        if exc_info is not None and not isinstance(exc_info, bool):
            raise ValueError('set_exception: exc_info must be bool')

        if exc:
            self._exc_info = (exc.__class__, str(exc), exc.__traceback__)
        elif exc_info == True:
            self._exc_info = sys.exc_info()

    def set_data(self, name: str, obj: Any) -> None:
        if self._data is None:
            self._data = {}

        if len(self._data) > TraceSpan.MAX_DATA_OBJECTS:
            raise ValueError('set_data: too many data objects (>{0})'.format(TraceSpan.MAX_DATA_OBJECTS))

        if name and not isinstance(name, str):
            raise ValueError('set_data: name must be string')

        self._data[name] = obj

    def _add_profiler_exception(self, exc):
        profiler_error = self._signal.profiler_errors.add()
        profiler_error.message = str(exc)
        if exc.__traceback__:
            frames = traceback.format_tb(exc.__traceback__)
            if len(frames) > 0:
                profiler_error.stack_trace = ''.join(frames)


class Tracer:
    def __init__(self, profiler=None):
        self._profiler = profiler

    def profiler(self):
        return self._profiler

    def span(self,
            endpoint: str,
            tags: Optional[Dict[str, str]] = None,
            ensure_trace: Optional[bool] = False) -> TraceSpan:

        return TraceSpan(
            endpoint=endpoint,
            tags=tags,
            ensure_trace=ensure_trace,
            operation_profiler=self._profiler)

def _timestamp_us():
    return int(time.time() * 1e6)



import unittest
import logging
import sys
from unittest.mock import patch, Mock
import torch
from google.protobuf.json_format import MessageToJson
import pprint
import time

import graphsignal
from graphsignal.usage.process_reader import ProcessReader
from graphsignal.proto import signals_pb2

logger = logging.getLogger('graphsignal')


class ProcessReaderTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_read(self):
        signal = signals_pb2.WorkerSignal()
        reader = ProcessReader()
        reader.read(signal)
        ProcessReader.MIN_CPU_READ_INTERVAL_NS = 0
        time.sleep(0.2)
        reader.read(signal)

        #pp = pprint.PrettyPrinter()
        # pp.pprint(MessageToJson(signal))

        self.assertNotEqual(signal.node_usage.hostname, '')
        self.assertNotEqual(signal.node_usage.ip_address, '')
        self.assertNotEqual(signal.process_usage.process_id, '')
        self.assertTrue(signal.process_usage.start_ms > 0)
        if sys.platform != 'win32':
            self.assertTrue(signal.node_usage.mem_total > 0)
            self.assertTrue(signal.node_usage.mem_used > 0)
            self.assertTrue(signal.process_usage.cpu_usage_percent > 0)
            self.assertTrue(signal.process_usage.max_rss > 0)
            self.assertTrue(signal.process_usage.current_rss > 0)
            self.assertTrue(signal.process_usage.vm_size > 0)
        self.assertTrue(signal.agent_info.version.major > 0 or signal.agent_info.version.minor > 0)



import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.data.builtin_types import BuiltInTypesProfiler

logger = logging.getLogger('graphsignal')


class BuiltInTypesProfilerTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_is_instance(self):
        profiler = BuiltInTypesProfiler()
        self.assertTrue(profiler.is_instance([]))

    def test_compute_counts(self):
        profiler = BuiltInTypesProfiler()
        self.assertEqual(
            profiler.compute_counts(dict(a=[0, 1, 1, 2.0, 3, float('nan'), float('inf')], b=[0, None, None], c=set([6,7]))), 
            {'element_count': 12, 'inf_count': 1, 'nan_count': 1, 'null_count': 2, 'zero_count': 2})

    def test_compute_counts_none(self):
        profiler = BuiltInTypesProfiler()
        self.assertEqual(
            profiler.compute_counts(None), 
            {'element_count': 1, 'inf_count': 0, 'nan_count': 0, 'null_count': 1, 'zero_count': 0})

    def test_build_stats(self):
        profiler = BuiltInTypesProfiler()
        ds = profiler.build_stats([1, 2])
        self.assertEqual(ds.data_type, 'list')

    def test_build_stats_none(self):
        profiler = BuiltInTypesProfiler()
        ds = profiler.build_stats(None)
        self.assertEqual(ds.data_type, 'NoneType')


import os
import time
import json
import gzip
import sys
import threading
import base64
import logging
from io import BytesIO
from urllib.request import urlopen
from urllib.request import Request
from urllib.parse import urlencode
from urllib.error import URLError
from urllib.error import HTTPError

import graphsignal
from graphsignal.proto import signals_pb2

logger = logging.getLogger('graphsignal')


class Uploader:
    MAX_BUFFER_SIZE = 25

    def __init__(self):
        self.agent_api_url = 'https://agent-api.graphsignal.com'
        self.buffer = []
        self.buffer_lock = threading.Lock()
        self.flush_lock = threading.Lock()

    def setup(self):
        if 'GRAPHSIGNAL_AGENT_API_URL' in os.environ:
            self.agent_api_url = os.environ['GRAPHSIGNAL_AGENT_API_URL']

    def clear(self):
        with self.buffer_lock:
            self.buffer = []

    def upload_signal(self, signal):
        with self.buffer_lock:
            self.buffer.append(signal)
            if len(self.buffer) > self.MAX_BUFFER_SIZE:
                self.buffer = self.buffer[-self.MAX_BUFFER_SIZE:]

    def flush_in_thread(self):
        threading.Thread(target=self.flush).start()

    def flush(self):
        with self.flush_lock:
            with self.buffer_lock:
                if len(self.buffer) == 0:
                    return
                outgoing = self.buffer
                self.buffer = []
            try:
                upload_start = time.time()
                payload = _create_upload_request(outgoing)
                content = self._post('signals', payload)
                _create_upload_response(content)
                logger.debug('Upload took %.3f sec (%dB)', time.time() - upload_start, len(payload))
            except URLError:
                logger.debug('Failed uploading signals, will retry', exc_info=True)
                with self.buffer_lock:
                    self.buffer[:0] = outgoing
            except Exception:
                logger.error('Error uploading signals', exc_info=True)

    def _post(self, endpoint, data):
        logger.debug('Posting data to %s/%s', self.agent_api_url, endpoint)

        api_key_64 = _base64_encode(graphsignal._agent.api_key + ':').replace('\n', '')
        headers = {
            'Accept-Encoding': 'gzip',
            'Authorization': "Basic %s" % api_key_64,
            'Content-Type': 'application/octet-stream',
            'Content-Encoding': 'gzip'
        }

        data_gzip = _gzip_data(data)

        request = Request(
            url=self.agent_api_url + '/' + endpoint,
            data=data_gzip,
            headers=headers)

        try:
            resp = urlopen(request, timeout=10)
            content = resp.read()
            if resp.info():
                content_type = resp.info().get('Content-Encoding')
                if content_type == 'gzip':
                    content = _gunzip_data(content)
            resp.close()
            return content
        except HTTPError as herr:
            logger.debug('Error message from server: %s',
                         herr.read().decode('utf-8'))
            raise herr


def _create_upload_request(outgoing):
    upload_request = signals_pb2.UploadRequest()
    upload_request.worker_signals.extend(outgoing)
    upload_request.upload_ms = int(time.time() * 1e3)
    return upload_request.SerializeToString()


def _create_upload_response(content):
    upload_response = signals_pb2.UploadResponse()
    upload_response.ParseFromString(content)
    return upload_response


def _gzip_data(data):
    gzip_out = BytesIO()
    with gzip.GzipFile(fileobj=gzip_out, mode="w") as out_file:
        out_file.write(data)
        out_file.close()

    gzip_out_val = gzip_out.getvalue()
    if isinstance(gzip_out_val, str):
        data_gzip = bytearray(gzip_out.getvalue())
    else:
        data_gzip = gzip_out.getvalue()

    return data_gzip


def _gunzip_data(data):
    return gzip.GzipFile('', 'r', 0, BytesIO(data)).read()


def _base64_encode(s):
    return base64.b64encode(s.encode('utf-8')).decode('utf-8')



from typing import Any, Union, Optional
import os
import logging
import atexit

from graphsignal.version import __version__
from graphsignal.agent import Agent

logger = logging.getLogger('graphsignal')

_agent = None


def _check_configured():
    global _agent
    if not _agent:
        raise ValueError(
            'Agent not configured, call graphsignal.configure() first')


def _check_and_set_arg(
        name, value, is_str=False, is_int=False, is_bool=False, required=False, max_len=None):
    env_name = 'GRAPHSIGNAL_{0}'.format(name.upper())

    if not value and env_name in os.environ:
        value = os.environ[env_name]
        if value:
            if is_int:
                try:
                    value = int(value)
                except:
                    raise ValueError('configure: invalid format, expected integer: {0}'.format(name))
            elif is_bool:
                value = bool(value)

    if not value:
        if required:
            raise ValueError('configure: missing argument: {0}'.format(name))
    else:
        if is_str:
            if not isinstance(value, str):
                raise ValueError('configure: invalid format, expected string: {0}'.format(name))
            if max_len and len(value) > max_len:
                raise ValueError('configure: invalid format, string too long (>{0}): {1}'.format(name, max_len))
        elif is_int:
            if not isinstance(value, int):
                raise ValueError('configure: invalid format, expected integer: {0}'.format(name))
        elif is_bool:
            if not isinstance(value, bool):
                raise ValueError('configure: invalid format, expected boolean: {0}'.format(name))

    return value


def configure(
        api_key: Optional[str] = None,
        debug_mode: Optional[bool] = False) -> None:
    global _agent

    if _agent:
        logger.warning('Agent already configured')
        return

    debug_mode = _check_and_set_arg('debug_mode', debug_mode, is_bool=True)
    if debug_mode:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.WARNING)
    api_key = _check_and_set_arg('api_key', api_key, is_str=True, required=True)

    _agent = Agent(
        api_key=api_key, 
        debug_mode=debug_mode)
    _agent.start()

    atexit.register(shutdown)

    logger.debug('Agent configured')


def tracer(with_profiler: Union[bool, str] = True):
    _check_configured()

    return _agent.tracer(with_profiler=with_profiler)


def upload(block=False) -> None:
    _check_configured()

    _agent.upload(block=block)


def shutdown() -> None:
    global _agent
    _check_configured()

    atexit.unregister(shutdown)
    _agent.stop()
    _agent = None

    logger.debug('Agent shutdown')


__all__ = [
    '__version__',
    'configure',
    'tracer'
    'upload',
    'shutdown'
]


import logging

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.data.data_profiler import DataProfiler, add_counts

logger = logging.getLogger('graphsignal')


class BuiltInTypesProfiler(DataProfiler):
    def __init__(self):
        super().__init__()

    def is_instance(self, data):
        return isinstance(data, (list, dict, tuple, set, str, int, float, complex)) or data is None

    def compute_counts(self, data):
        counts = {}
        if isinstance(data, str):
            counts['char_count'] = len(data)
        elif isinstance(data, bytes):
            counts['byte_count'] = len(data)
        else:
            counts['element_count'] = _count_elems(data, lambda elem: True)
        counts['null_count'] = _count_elems(data, lambda elem: elem is None)
        counts['nan_count'] = _count_elems(data, lambda elem: elem != elem)
        counts['inf_count'] = _count_elems(data, lambda elem: elem == float("inf"))
        counts['zero_count'] = _count_elems(data, lambda elem: elem == 0)
        return counts

    def build_stats(self, data):
        counts = self.compute_counts(data)
        data_stats = signals_pb2.DataStats()
        data_stats.data_type = type(data).__name__
        add_counts(data_stats, counts)
        return data_stats


def _count_elems(data, filter_func):
    if isinstance(data, dict):
        return sum([_count_elems(elem, filter_func) for elem in data.values()])
    elif isinstance(data, (list, tuple, set)):
        return sum([_count_elems(elem, filter_func) for elem in data])
    elif isinstance(data, (int, float, complex, str, bytes)):
        return sum([filter_func(data)])
    elif data is None:
        return sum([filter_func(data)])
    return 0
